/*
 *  TOPPERS/ASP Kernel
 *      Toyohashi Open Platform for Embedded Real-Time Systems/
 *      Advanced Standard Profile Kernel
 * 
 *  Copyright (C) 2000-2003 by Embedded and Real-Time Systems Laboratory
 *                              Toyohashi Univ. of Technology, JAPAN
 *  Copyright (C) 2003-2004 by Naoki Saito
 *             Nagoya Municipal Industrial Research Institute, JAPAN
 *  Copyright (C) 2003-2004 by Platform Development Center
 *                                          RICOH COMPANY,LTD. JAPAN
 *  Copyright (C) 2008-2010 by Witz Corporation, JAPAN
 *  Copyright (C) 2013      by Mitsuhiro Matsuura
 * 
 *  上記著作権者は，以下の(1)～(4)の条件を満たす場合に限り，本ソフトウェ
 *  ア（本ソフトウェアを改変したものを含む．以下同じ）を使用・複製・改
 *  変・再配布（以下，利用と呼ぶ）することを無償で許諾する．
 *  (1) 本ソフトウェアをソースコードの形で利用する場合には，上記の著作
 *      権表示，この利用条件および下記の無保証規定が，そのままの形でソー
 *      スコード中に含まれていること．
 *  (2) 本ソフトウェアを，ライブラリ形式など，他のソフトウェア開発に使
 *      用できる形で再配布する場合には，再配布に伴うドキュメント（利用
 *      者マニュアルなど）に，上記の著作権表示，この利用条件および下記
 *      の無保証規定を掲載すること．
 *  (3) 本ソフトウェアを，機器に組み込むなど，他のソフトウェア開発に使
 *      用できない形で再配布する場合には，次のいずれかの条件を満たすこ
 *      と．
 *    (a) 再配布に伴うドキュメント（利用者マニュアルなど）に，上記の著
 *        作権表示，この利用条件および下記の無保証規定を掲載すること．
 *    (b) 再配布の形態を，別に定める方法によって，TOPPERSプロジェクトに
 *        報告すること．
 *  (4) 本ソフトウェアの利用により直接的または間接的に生じるいかなる損
 *      害からも，上記著作権者およびTOPPERSプロジェクトを免責すること．
 *      また，本ソフトウェアのユーザまたはエンドユーザからのいかなる理
 *      由に基づく請求からも，上記著作権者およびTOPPERSプロジェクトを
 *      免責すること．
 * 
 *  本ソフトウェアは，無保証で提供されているものである．上記著作権者お
 *  よびTOPPERSプロジェクトは，本ソフトウェアに関して，特定の使用目的
 *  に対する適合性も含めて，いかなる保証も行わない．また，本ソフトウェ
 *  アの利用により直接的または間接的に生じたいかなる損害に関しても，そ
 *  の責任を負わない．
 * 
 *  @(#) $Id: target_timer.c 313 2017-07-23 04:50:32Z coas-nagasima $
 */

/*
 *		タイマドライバ
 */

#include "kernel_impl.h"
#include "time_event.h"
#include <sil.h>
#include "target_timer.h"

/*
 *  タイマの初期化処理
 */
void
target_timer_initialize(intptr_t exinf)
{
	/*
	 * モジュールストップ機能の設定
	 */
	sil_wrh_mem((void *)SYSTEM_PRCR_ADDR, (uint16_t)0xA502);	/* 書込み許可 */
	*SYSTEM_MSTPCRA_ADDR &= ~(SYSTEM_MSTPCRA_MSTPA15_BIT);
	sil_wrh_mem((void *)SYSTEM_PRCR_ADDR, (uint16_t)0xA500);	/* 書込み禁止 */

	/* 
	 * タイマ停止 
	 */
	*CMT_CMSTR0_ADDR &= ~CMT_CMSTR0_STR0_BIT;
	
	/*
 	 * カウントアップに用いられるクロック設定
	 * PCLK/32を選択
	 */
	*CMT0_CMCR_ADDR = CMT_PCLK_DIV_8;

	/* 
	 * コンペアマッチタイマカウンタ設定
	 */
	*CMT0_CMCNT_ADDR = 0U;

	/* 
	 * コンペアマッチタイマ周期設定
	 */
	*CMT0_CMCOR_ADDR = CMCOR_PERIOD;

	/*
	 * コンペアマッチタイマ割り込み要求先設定レジスタ（ISELR28）
	 */
	*ICU_ISELR028_ADDR = ICU_ISEL_CPU;

	/*
	 *  タイマ動作開始前の割込み要求をクリア
	 */
	x_clear_int(INTNO_TIMER);

	/* 
	 * コンペアマッチタイマ割り込みを許可
	 */
	*CMT0_CMCR_ADDR |= CMT0_CMCR_CMIE_BIT;
	
	/* 
	 * タイマ動作開始
	 */
	*CMT_CMSTR0_ADDR |= CMT_CMSTR0_STR0_BIT;
}

/*
 *  タイマの停止処理
 */
void
target_timer_terminate(intptr_t exinf)
{
	/* 
	 * タイマ停止 
	 */
	*CMT_CMSTR0_ADDR &= ~CMT_CMSTR0_STR0_BIT;
}

/*
 *  タイマ割込みハンドラ
 */
void target_timer_handler(void)
{
	i_begin_int(INTNO_TIMER);
	
	signal_time();				/* タイムティックの供給 */
	
	i_end_int(INTNO_TIMER);
}
